/*
 * Copyright 2010 James Geboski <jgeboski@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "libg19draw.h"

#ifdef FT2_SUPPORT
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_BITMAP_H

static FT_Library ft_lib;
static FT_Face ft_face;
static int ft_width;
static int ft_height;
static int ft_space;
static unsigned int ft_color;
#endif

/**
 * Creates a new G19Draw
 * 
 * @param width width in pixels of image to create
 * @param height height in pixels of image to create
 * 
 * @return pointer to a G19Draw
 **/
G19Draw * g19_draw_new(int width, int height)
{
	G19Draw * draw;
	
	draw = (G19Draw *) malloc(sizeof(G19Draw));
	
	draw -> width = width;
	draw -> height = height;
	draw -> size = width * height * 4;
	
	draw -> data = (unsigned char *) malloc(draw -> size);
	memset(draw -> data, 0x00, draw -> size);
	
	return draw;
}

/**
 * Creates a new G19Draw from an image
 * The image must me a 24-bit or 32-bit bitmap with no compression
 * 
 * @param path path to the image to load
 * 
 * @return pointer to a G19Draw; On failure NULL
 **/
G19Draw * g19_draw_bmp(char * path)
{
	G19Draw * draw;
	FILE * file;
	int bits, comp;
	int width, height;
	unsigned int doff;
	
	if((file = fopen(path, "rb")) == NULL)
		return NULL;
	
	bits = 0;
	fseek(file, 0x1c, SEEK_SET);
	fread(&bits, 2, 1, file);
	
	if((bits != 24) && (bits != 32))
	{
		fclose(file);
		return NULL;
	}
	
	comp = -1;
	fseek(file, 0x1e, SEEK_SET);
	fread(&comp, 4, 1, file);
	
	if((comp < 0) || (comp != 0))
	{
		fclose(file);
		return NULL;
	}
	
	width = 0;
	fseek(file, 0x12, SEEK_SET);
	fread(&width, 4, 1, file);
	
	if(width <= 0)
	{
		fclose(file);
		return NULL;
	}
	
	height = 0;
	fseek(file, 0x16, SEEK_SET);
	fread(&height, 4, 1, file);
	
	if(height <= 0)
	{
		fclose(file);
		return NULL;
	}
	
	doff = 0;
	fseek(file, 0xa, SEEK_SET);
	fread(&doff, 4, 1, file);
	
	if(doff == 0)
	{
		fclose(file);
		return NULL;
	}
	
	draw = g19_draw_new(width, height);
	if(draw == NULL)
	{
		fclose(file);
		return NULL;
	}
	
	int x, y;
	unsigned char buffer[4];
	unsigned int color;
	
	fseek(file, doff, SEEK_SET);
	
	for(x = 0; x < width; x++)
	{
		for(y = 0; y < height; y++)
		{
			if(bits == 24)
			{
				fread(buffer, 3, 1, file);
				color = G19Color(buffer[2], buffer[1], buffer[0]);
			}
			else if(bits == 32)
			{
				fread(buffer, 4, 1, file);
				color = G19ColorA(buffer[2], buffer[1], buffer[0], buffer[3]);
			}
			
			g19_draw_set_pixel(draw, y + (width - height), ((width + (height - width)) - 1) - x, color);
		}
	}
	
	fclose(file);
	
	return draw;
}

/**
 * Frees up up allocated memory the G19Draw used
 * 
 * @param draw pointer to the G19Draw to free
 * 
 * @return pointer to a G19Draw
 **/
void g19_draw_free(G19Draw * draw)
{
	if(draw -> data != NULL)
		free(draw -> data);
	
	if(draw != NULL)
		free(draw);
}

/**
 * Insert a G19Draw onto another G19Draw
 * 
 * @param draw pointer to the destination G19Draw
 * @param idraw pointer to the source G19Draw
 * @param ix x coordinate of the upper-left
 * @param iy y coordinate of the upper-left
 **/
void g19_draw_insert(G19Draw * draw, G19Draw * idraw, int ix, int iy)
{
	if((draw == NULL) || (idraw == NULL))
		return;
	
	if((ix > draw -> width) || (iy > draw -> height))
		return;
	
	int ex, ey;
	int x, y;
	int dd, id;
	
	ex = ((ix + idraw -> width) < draw -> width) ? idraw -> width : (draw -> width - ix);
	ey = ((iy + idraw -> height) < draw -> height) ? idraw -> height : (draw -> height - iy);
	
	for(x = 0; x < ex; x++)
	{
		for(y = 0; y < ey; y++)
		{
			dd = ((draw -> height * 4) * (ix + x)) + ((iy + y) * 4);
			id = ((idraw -> height * 4) * x) + (y * 4);
			
			memcpy(draw -> data + dd, idraw -> data + id, 4);
		}
	}
}

/**
 * Insert a G19Draw onto another G19Draw with alpha blending
 * 
 * @param draw pointer to the destination G19Draw
 * @param idraw pointer to the source G19Draw
 * @param ix x coordinate of the upper-left
 * @param iy y coordinate of the upper-left
 * @param opac if less that 0 then it will use the values stores or specify 0 - 255
 **/
void g19_draw_insert_ab(G19Draw * draw, G19Draw * idraw, int ix, int iy, int opac)
{
	if((draw == NULL) || (idraw == NULL))
		return;
	
	if((ix > draw -> width) || (iy > draw -> height))
		return;
	
	if(opac > 256)
		opac = 256;
	
	int ex, ey;
	int x, y;
	int dd, id;
	unsigned char d1[4], d2[4];
	int r, g, b;
	unsigned int color;
	
	ex = ((ix + idraw -> width) < draw -> width) ? idraw -> width : (draw -> width - ix);
	ey = ((iy + idraw -> height) < draw -> height) ? idraw -> height : (draw -> height - iy);
	
	for(x = 0; x < ex; x++)
	{
		for(y = 0; y < ey; y++)
		{
			dd = ((draw -> height * 4) * (ix + x)) + ((iy + y) * 4);
			id = ((idraw -> height * 4) * x) + (y * 4);
			
			memcpy(&d1, draw -> data + dd, 4);
			memcpy(&d2, idraw -> data + id, 4);
			
			r = d1[0] + (((d2[0] - d1[0]) * ((opac >= 0) ? opac : d2[3])) / 256);
			g = d1[1] + (((d2[1] - d1[1]) * ((opac >= 0) ? opac : d2[3])) / 256);
			b = d1[2] + (((d2[2] - d1[2]) * ((opac >= 0) ? opac : d2[3])) / 256);
			
			color = G19Color(r, g, b);
			memcpy(draw -> data + dd, &color, 4);
		}
	}
}

/**
 * Gets the pixel value at the specified coordinates
 * 
 * @param draw a pointer G19Draw
 * @param x x coordinate of pixel
 * @param y y coordinate of pixel
 * 
 * @return an unsigned int that can be used in unison with: G19R(), G19G(), G19B(), G19A()
 **/
unsigned int g19_draw_get_pixel(G19Draw * draw, int x, int y)
{
	unsigned int pixel;
	
	int pos = ((draw -> height * 4) * x) + (y * 4);
	memcpy(&pixel, draw -> data + pos, 4);
	
	return pixel;
}

/**
 * Sets the pixel value at the specified coordinates
 * 
 * @param draw a pointer G19Draw
 * @param x x coordinate of pixel
 * @param y y coordinate of pixel
 * @param color an unsigned int that hold the color; G19Color() or G19ColorA() can be used
 **/
void g19_draw_set_pixel(G19Draw * draw, int x, int y, unsigned int color)
{
	if(draw == NULL)
		return;
	
	if((x < 0) || (x > draw -> width))
		return;
	
	if((y < 0) || (y > draw -> height))
		return;
	
	int pos = ((draw -> height * 4) * x) + (y * 4);
	
	memcpy(draw -> data + pos, &color, 4);
}

/**
 * Fills the G19Draw with a color
 * 
 * @param draw a pointer G19Draw
 * @param color an unsigned int that hold the color; G19Color() or G19ColorA() can be used
 **/
void g19_draw_fill(G19Draw * draw, unsigned int color)
{
	if(draw == NULL)
		return;
	
	int i;
	
	for(i = 0; i < (draw -> size / 4); i++)
		memcpy(draw -> data + (i * 4), &color, 4);
}

/**
 * Draws a line on the G19Draw
 * 
 * @param draw a pointer G19Draw
 * @param x1 x coordinate of the first point
 * @param y1 y coordinate of the first point
 * @param x2 x coordinate of the second point
 * @param y2 y coordinate of the second point
 * @param color an unsigned int that hold the color; G19Color() or G19ColorA() can be used
 **/
void g19_draw_line(G19Draw * draw, int x1, int y1, int x2, int y2, unsigned int color)
{
	if(draw == NULL)
		return;
	
	//Bresenham's Line Algorithm
	
	int steep, hold;
	int dx, dy, err;
	int ystep, x, y;
	
	steep = (abs(y2 - y1) > abs(x2 - x1)) ? 1 : 0;
	
	if(steep)
	{
		hold = x1;
		x1 = y1;
		y1 = hold;
		
		hold = x2;
		x2 = y2;
		y2 = hold;
	}
	
	if(x1 > x2)
	{
		hold = x1;
		x1 = x2;
		x2 = hold;
		
		hold = y1;
		y1 = y2;
		y2 = hold;
	}
	
	dx = x2 - x1;
	dy = abs(y2 - y1);
	
	err = dx / 2;
	y = y1;
	ystep = (y1 < y2) ? 1 : 0;
	
	for(x = x1; x < x2; x++)
	{
		if(steep)
			g19_draw_set_pixel(draw, y, x, color);
		else
			g19_draw_set_pixel(draw, x, y, color);
		
		err = err - dy;
		if(!err)
		{
			y = y + ystep;
			err = err + dx;
		}
	}
}

/**
 * Draws a rectangle on the G19Draw
 * 
 * @param draw a pointer G19Draw
 * @param left left most point int terms of x
 * @param top top most point int terms of y
 * @param right most point int terms of x
 * @param bottom most point int terms of y
 * @param fill 0 for 1 pixel border or 1 to fill the rectangle
 * @param color an unsigned int that hold the color; G19Color() or G19ColorA() can be used
 **/
void g19_draw_rect(G19Draw * draw, int left, int top, int right, int bottom, int fill, unsigned int color)
{
	if(draw == NULL)
		return;
	
	if(!fill)
	{
		g19_draw_line(draw, left, top, right, top, color);
		g19_draw_line(draw, left, bottom, right, bottom, color);
		g19_draw_line(draw, left, top, left, bottom, color);
		g19_draw_line(draw, right, top, right, bottom, color);
		
		return;
	}
	
	int y, x;
	
	for(x = left; x < right; x++)
	{
		for(y = top; y < bottom; y++)
			g19_draw_set_pixel(draw, x, y, color);
	}
}

/**
 * Draws a circle on the G19Draw
 * 
 * @param draw a pointer G19Draw
 * @param cx x coordinate of the circle's center
 * @param cy y coordinate of the circle's center
 * @param cr radius of the circle
 * @param fill 0 for 1 pixel border or 1 to fill the rectangle
 * @param color an unsigned int that hold the color; G19Color() or G19ColorA() can be used
 **/
void g19_draw_circle(G19Draw * draw, int cx, int cy, int cr, int fill, unsigned int color)
{
	if(draw == NULL)
		return;
	
	//Bresenham's circle algorithm
	
	int err = -cr;
	int x = cr;
	int y = 0;
	
	while(x >= y)
	{
		if(fill)
		{
			g19_draw_line(draw, cx + x, cy + y, cx + x, cy - y, color);
			g19_draw_line(draw, cx - x, cy - y, cx - x, cy + y, color);
			
			if(x != y)
			{
				g19_draw_line(draw, cx + y, cy + x, cx + y, cy - x, color);
				g19_draw_line(draw, cx - y, cy - x, cx - y, cy + x, color);
			}
		}
		else
		{
			g19_draw_set_pixel(draw, cx + x, cy + y, color);
			g19_draw_set_pixel(draw, cx - x, cy + y, color);
			g19_draw_set_pixel(draw, cx + x, cy - y, color);
			g19_draw_set_pixel(draw, cx - x, cy - y, color);
			
			if(x != y)
			{
				g19_draw_set_pixel(draw, cx + y, cy + x, color);
				g19_draw_set_pixel(draw, cx - y, cy + x, color);
				g19_draw_set_pixel(draw, cx + y, cy - x, color);
				g19_draw_set_pixel(draw, cx - y, cy - x, color);
			}
		}
		
		err += y;
		y++;
		err += y;
		
		if(err >= 0)
		{
			x--;
			err -= x;
			err -= x;
		}
	}
	
}

/**
 * Draws a slot from FreeType on the G19Draw
 * 
 * @param draw a pointer G19Draw
 * @param fslot G19FontSlot that holds the FreeType slot data
 **/
void g19_draw_ft_slot(G19Draw * draw, G19FontSlot * fslot, int x, int y)
{
	int xx, yy;
	unsigned char tc;
	unsigned int fcolor;
	
	if(draw == NULL)
		return;
	
	G19FontBitmap * bitmap = fslot -> bitmap;
	
	G19Draw * tdraw = g19_draw_new(fslot -> width, fslot -> bitmap_top + fslot -> height);
	g19_draw_fill(tdraw, G19ColorA(0, 0, 0, 0));
	
	if(tdraw == NULL)
		return;
	
	for(yy = 0; yy < bitmap -> rows; yy++)
	{
		for(xx = 0; xx < bitmap -> width; xx++)
		{
			tc = bitmap -> buffer[(yy * bitmap -> width) + xx];
			
			if(tc == 0)
				continue;
			
			fcolor = G19ColorA(G19R(fslot -> color), G19G(fslot -> color),
							  G19B(fslot -> color), (256 + tc));
			
			g19_draw_set_pixel(tdraw, fslot -> bitmap_left + xx, (fslot -> height - fslot -> bitmap_top) + yy, fcolor);
		}
	}
	
	g19_draw_insert_ab(draw, tdraw, x, y, -1);
	g19_draw_free(tdraw);
}

#ifdef FT2_SUPPORT

/**
 * Sets the font for drawing
 * 
 * @param ttf_path path to the ttf file
 * @param width width of text
 * @param height height of text
 * @param space space between chars in pixels
 * @param color an unsigned int that hold the color; G19Color() or G19ColorA() can be used
 * 
 * @return non zero on failure
 **/
int g19_draw_set_font(char * ttf_path, int width, int height, int space, unsigned int color)
{
	int res = FT_Init_FreeType(&ft_lib);
	if(res)
		return res;
	
	res = FT_New_Face(ft_lib, ttf_path, 0, &ft_face);
	if(res)
		return res;
	
	res = FT_Set_Pixel_Sizes(ft_face, width, height);
	if(res)
		return res;
	
	ft_width = width;
	ft_height = height;
	ft_space = space;
	ft_color = color;
	
	if(space < 0)
		space = 0;
	
	return 0;
}

/**
 * Draws text on the G19Draw
 * Make sure you have executed g19_draw_set_font
 * 
 * @param draw a pointer G19Draw
 * @param x x coordinate of the upper-left
 * @param y y coordinate of the upper-left
 * @param text pointer to an array of chars to be drawn
 * 
 * @return non zero on failure
 **/
int g19_draw_text(G19Draw * draw, int x, int y, char * text)
{
	int res, i;
	FT_GlyphSlot slot;
	FT_UInt gi;
	
	if((ft_lib == NULL) || (ft_face == NULL))
		return -1;
	
	slot = ft_face -> glyph;
	
	for(i = 0; i < strlen(text); i++)
	{
		gi = FT_Get_Char_Index(ft_face, text[i]);
		
		if(gi == 0)
			continue;
		
		res = FT_Load_Char(ft_face, (FT_ULong) text[i], FT_LOAD_DEFAULT);
		if(res)
			continue;
		
		res = FT_Render_Glyph(ft_face -> glyph, FT_RENDER_MODE_NORMAL);
		if(res)
			continue;
		
		if(slot -> bitmap.width < 1)
		{
			x += ft_width;
			continue;
		}
		
		x += ft_space;
		
		G19FontSlot fslot;
		fslot.width = ft_width;
		fslot.height = ft_height;
		fslot.bitmap_left = slot -> bitmap_left;
		fslot.bitmap_top = slot -> bitmap_top;
		fslot.color = ft_color;
		fslot.bitmap = (G19FontBitmap *) &slot -> bitmap;
		
		g19_draw_ft_slot(draw, &fslot, x, y);
		
		x += slot -> bitmap.width;
	}
	
	return 0;
}

/**
 * Closes the FreeType library and font handle from g19_draw_set_font()
 **/
void g19_draw_close_font(void)
{
	if(ft_face != NULL)
		FT_Done_Face(ft_face);
	
	if(ft_lib != NULL)
		FT_Done_FreeType(ft_lib);
}

#endif
